///////////////////////////////////////////////
//                                           //
// VRLab Mental Vision Module 5: NURBS       //
// Author: Achille Peternier                 //
// Copyright: EPFL-VRLab 2005 (C)            //
//                                           //
///////////////////////////////////////////////
#include "../data/common.h"
#include "bezier.h"
#pragma comment(lib, "opengl32.lib")



//////////////
// #DEFINES //
//////////////
#define MODULE_NAME "NURBS surfaces"



//////////////
// ACCURACY //
//////////////
#define ACCURACY 20  // Lines per fragment
#define POINTSX  4    
#define POINTSY  4



//////////
// MAIN //
//////////

int main(int argc, char *argv[])
{ 
   /////////////////
   // Init MVisio //
   /////////////////
   INIT_MVISIO(MODULE_NAME, argc, argv);   
   
   // Common interface:
   USE_FONT();  
   USE_BACKGROUND(MODULE_SIZEX, MODULE_SIZEY);
   USE_QUITBUTTON(MODULE_SIZEX, MODULE_SIZEY);
   USE_CAMERA(0.0f, 0.0f, 190.0f, 0.0f, 0.0f, 0.0f);    
   USE_LIGHT(50.0f, 100.0f, 0.0f);
   USE_SELECTBOX();  
   USE_SELECTTORUS();
  
   
   /////////////
   // Patches //
   /////////////       
   
   patch_bezier patch;
	EXF_ZEROMEMORY(&patch, sizeof(patch_bezier));
	patch.affiche_pt_ctrl = true;
	patch.affiche_vect_derive = true;		
	
	// Feed the input struct:
   Grille_triplet bezier;
	bezier.nb_colonnes = POINTSX;
	bezier.nb_lignes = POINTSY;
	bezier.grille = (Triplet **) malloc(sizeof(Triplet *) * bezier.nb_lignes);
   for (int c=0; c<bezier.nb_lignes; c++)
         bezier.grille[c] = (Triplet *) malloc(sizeof(Triplet) * bezier.nb_colonnes);
         
   // Feed data:
   float SIZE = 40.0f; 
   for (int y=0; y<POINTSY; y++)
	      for (int x=0; x<POINTSX; x++)
		   {
		         bezier.grille[x][y].x = x*SIZE - (SIZE * POINTSX)/2.0f+SIZE/2;
		         bezier.grille[x][y].y = 0.0f;
		         bezier.grille[x][y].z = y*SIZE - (SIZE * POINTSY)/2.0f+SIZE/2;
		   }   
		   
   
   ////////////////////////////
   // Point modifier interface:
   MVWINDOW *interfaceWin = new MVWINDOW();
   interfaceWin->setFont(font);
   interfaceWin->setSizes(220, 110);  
   interfaceWin->setCaption("Point modifier");
   interfaceWin->setTexture(baseTexture);
   interfaceWin->setPosition((float) (MODULE_SIZEX-MODULE_SIZEX/3), (float) (MODULE_SIZEY-MODULE_SIZEY/3));		      
   
   MVBUTTON *randomizeButton = new MVBUTTON();
   randomizeButton->setFont(font);
   randomizeButton->setCaption("random");
   randomizeButton->setSizes(48, 20);
   randomizeButton->setPosition(10, 80);
   randomizeButton->link(interfaceWin);         
   
   MVBUTTON *resetButton = new MVBUTTON();
   resetButton->setFont(font);
   resetButton->setCaption("reset");
   resetButton->setSizes(40, 20);
   resetButton->setPosition(65, 80);
   resetButton->link(interfaceWin); 
   
   MVBUTTON *hideButton = new MVBUTTON();
   hideButton->setFont(font);
   hideButton->setCaption("hide norm/lines");
   hideButton->setSizes(96, 20);
   hideButton->setPosition(110, 80);
   hideButton->setEventType(MV_BUTTON_SWITCH);
   hideButton->link(interfaceWin);    
   
   MVHSCROLLBAR *accuracyScroll = new MVHSCROLLBAR();
   accuracyScroll->setSizeX(104);
   accuracyScroll->setMinValue(2.0);
   accuracyScroll->setMaxValue(80.0f);
   accuracyScroll->setValue(ACCURACY);
   accuracyScroll->setPosition(100, 11);
   accuracyScroll->link(interfaceWin);   
   
   // Tension:
   MVHSCROLLBAR *modifyPointScroll1 = new MVHSCROLLBAR();   
   modifyPointScroll1->setPosition(120, 40);
   modifyPointScroll1->link(interfaceWin);
   modifyPointScroll1->setMinValue(-1.0f);
   modifyPointScroll1->setMaxValue(1.0f);
   modifyPointScroll1->setValue(0.0f);
   
   // Bias:
   MVHSCROLLBAR *modifyPointScroll2 = new MVHSCROLLBAR();   
   modifyPointScroll2->setPosition(120, 61);
   modifyPointScroll2->link(interfaceWin);
   modifyPointScroll2->setMinValue(-1.0f);   
   modifyPointScroll2->setMaxValue(1.0f);   
   modifyPointScroll2->setValue(0.0f);
   
  
   ////////////////
   // Module slide:
   MVTEXTURE *slideTexture = new MVTEXTURE();
   slideTexture->load("module5_slide.tga");
   MVWINDOW *slideWindow = new MVWINDOW();
   slideWindow->setFont(font);
   slideWindow->setCaption("Slide");
   slideWindow->setSizes(512, 440);
   slideWindow->setColor(1.0f, 1.0f, 1.0f, 1.0f);
   slideWindow->setTexture(slideTexture);
   slideWindow->setPosition(15, 85);
   
   MVVSCROLLBAR *slideScroll = new MVVSCROLLBAR();
   slideScroll->setMinValue(0.1f);
   slideScroll->setMaxValue(1.0f);
   slideScroll->setValue(1.0f);
   slideScroll->setPosition(495, 12);
   slideScroll->link(slideWindow);                 

///////////////
// Mainloop: //
///////////////

   int mouseX = 0, mouseY = 0, oldMouseX = 0, oldMouseY = 0;
   bool mouseBL = false, mouseBR = false;
   bool done = false;
   float rotationX = 40.0f, rotationY = 0.0f;
   MVID currentSelected = 0, locked = 0;
   MVID lastSelected = 0;
   bool moving = false;
   float selRot = 0.0f; 
   float zoom = 220.0f;
   while (!done)   
	{			
	      // Rendering:	   
	      camera->setPosition(0.0f, 0.0f, zoom);
	      selRot += 180.0f * MVISIO::getFPSfactor();
	   	MVISIO::clear(true, true, false);	   		   	
			MVISIO::begin2D();			   
			   PASS_BACKGROUND;					     			  
			MVISIO::end2D();				     
			
			// Draw Title:
			MVISIO::begin2D();					   
			   // Title:
			   font->putString(10, 13, MODULE_NAME);	  			   
			MVISIO::end2D();   
		   
			// Compute line:			
			patch.gr_cont = bezier;
			patch.nb_pt_u = (int) accuracyScroll->getValue();
			patch.nb_pt_v = (int) accuracyScroll->getValue();							
			changement(&patch);			
			MVISIO::begin3D(camera);				      	    
			      glRotatef(rotationX, 1.0f, 0.0f, 0.0f);
			glRotatef(rotationY, 0.0f, 1.0f, 0.0f);	
			      // Display control boxes:
			      light->pass();			      
			      for (int x=0; x<bezier.nb_colonnes; x++)
			         for (int y=0; y<bezier.nb_lignes; y++)
			         {
			               selectBox->setPosition(bezier.grille[x][y].x, bezier.grille[x][y].y, bezier.grille[x][y].z);
			               selectBox->getChild(0)->setAuxId(y*bezier.nb_colonnes+x+1);			               
			               selectBox->pass();
			               
			               // Current selected:
			               if (locked == y*bezier.nb_colonnes+x+1)
			               {
			                     selectTorus->getChild(0)->setAuxId(y*bezier.nb_colonnes+x+1);		
			                     selectTorus->setPosition(bezier.grille[x][y].x, bezier.grille[x][y].y, bezier.grille[x][y].z);
			                     selectTorus->setRotation(selRot, selRot/2.0f, 0.0f);
			                     selectTorus->pass();
			                     selectTorus->setRotation(selRot/2.0f, selRot, 90.0f);
			                     selectTorus->pass();
			               }
			         }			         			      
			MVISIO::end3D();	
			currentSelected = MVISIO::select(mouseX, mouseY);				
			MVISIO::begin3D(camera);				      	    
			      glRotatef(rotationX, 1.0f, 0.0f, 0.0f);
			      glRotatef(rotationY, 0.0f, 1.0f, 0.0f);				
			      affiche_patch_bezier(&patch, 1.0f);
			MVISIO::end3D();
			   
			// Draw post interface:
			char text[356];
			sprintf(text, "x0005y0005tesselation: %d x0003--------------------------\nx0005tension: x0009%.2f\nx0005y0005bias: x0025%.2f", (int) accuracyScroll->getValue(), modifyPointScroll1->getValue(), modifyPointScroll2->getValue());
			interfaceWin->clear();
			interfaceWin->write(text);
			interfaceWin->format();			
			MVISIO::begin2D();
			   PASS_QUITBUTTON;					   
			   interfaceWin->pass();	   			
			   slideWindow->setColor(1.0f, 1.0f, 1.0f, slideScroll->getValue());
			   slideWindow->pass();	  
			MVISIO::end2D();
			
			
			////////////////
			// Check events:			
			SDL_Event event;			
			//SDL_PumpEvents();						
			while (SDL_PollEvent(&event)) 
			{
					// Quit by click on X:
					if (event.type == SDL_QUIT) 
							done = true;		
               
               // Resize?
				   if (event.type == SDL_VIDEORESIZE)				   
				         MVSETTINGS::resize(0, 0, event.resize.w, event.resize.h);																		                      
               
               // Quit by pressing ESC:
					if (event.type == SDL_KEYDOWN) 
					{
							// Quit by ESC:
							if (event.key.keysym.sym == SDLK_ESCAPE) 							       
							       done = true;	                  								
			      }														
			}         
			
			// Update keyboard (like a joypad):
			Uint8 *array;
			int arraySize;
			array = SDL_GetKeyState(&arraySize);
			if (array[SDLK_PAGEUP])    
			{
               zoom += 100.0f * MVISIO::getFPSfactor();
               if (zoom > 350.0f) zoom = 350.0f;			            
         }							       								                     
         else
               if (array[SDLK_PAGEDOWN])
               {
                     zoom -= 100.0f * MVISIO::getFPSfactor();
                     if (zoom < 50.0f) zoom = 50.0f;			            
               }							      
			
			// Update mouse:
			int buttonState = SDL_GetMouseState(&mouseX, &mouseY);		
         if (buttonState&SDL_BUTTON(1))
					mouseBL = true;					
			else 
					mouseBL = false;
         if (buttonState&SDL_BUTTON(2) || buttonState&SDL_BUTTON(3))
					mouseBR = true;					
			else 
					mouseBR = false;					
					
         float deltaX = mouseX - oldMouseX;
         float deltaY = mouseY - oldMouseY;		
         oldMouseX = mouseX; oldMouseY = mouseY;			
			
			// Check MVisio events:
			MVID id=0;
			if (!MVELEMENT::manageGUIEvents(NULL, mouseX, mouseY, mouseBL, mouseBR, &id))
			{
			      // Zoom:
			      if (mouseBL && mouseBR && locked==0 && (currentSelected==0 || moving))
			      {
			            zoom += deltaY;
			            if (zoom > 350.0f) zoom = 350.0f;
			            if (zoom < 50.0f) zoom = 50.0f;
			            moving = true;
			      }
			      else
			            // Rotate scene:
			            if (mouseBL && locked==0 && (currentSelected==0 || moving))
			            {
			                  // Rotate model:
			                  rotationX += deltaY/4.0f;
			                  rotationY += deltaX/4.0f;
			                  if (rotationX > 68.0f) rotationX = 68.0f;
			                  if (rotationX < -68.0f) rotationX = -68.0f;
			                  moving = true;			            
			            }
			      
			      if (moving && mouseBL == 0)
			            moving = false;
			
			      // Move point:
			      if (currentSelected && mouseBL && locked==0 && !moving)
			      {
			            locked = currentSelected;			         
		        	}
		        	else
		        	      if (!mouseBL)
		        	            locked = 0;
		        	
		        	if (locked)
		        	{
		        	      for (int x=0; x<bezier.nb_colonnes; x++)
			               for (int y=0; y<bezier.nb_lignes; y++)			         
			                     if (locked == y*bezier.nb_colonnes+x+1)
			                     {
			                           MVMATRIX * mat = MVCAMERA::getCameraMatrix();
	                                 MVVECTOR3 vec_x(mat->d[0], mat->d[4], mat->d[8]);
	                                 MVVECTOR3 vec_y(mat->d[1], mat->d[5], mat->d[9]);
	                                 MVVECTOR3 vec_z(mat->d[2], mat->d[6], mat->d[10]);
	                                 
	                                 // Object rotation:
	                                 MVMATRIX rot;
	                                 rot.setIdentity();
	                                 rot.setRotationX(rotationX);
	                                 rot.setRotationY(-rotationY);	                                 
	                                 	                               
	                                 float mx = deltaX/2.0f;
	                                 float my = -deltaY/2.0f;
			                           MVVECTOR3 depl = vec_x * mx + vec_y * my;
			                           depl = rot.rotate(depl);
			                           bezier.grille[x][y].x += depl.x; 
			                           bezier.grille[x][y].y += depl.y;
			                           bezier.grille[x][y].z += depl.z;			                           
      			               }     
      		   }
		   }			
			
			// Handle buttons:
			if (PRESSED_QUITBUTTON(id))
			      done = true;
			      
			if (id == randomizeButton->getId())
			{
			      for (int y=0; y<POINTSY; y++)
	                  for (int x=0; x<POINTSX; x++)
		               {
		                     bezier.grille[x][y].x = x*SIZE - (SIZE * POINTSX)/2.0f+SIZE/2;
		                     bezier.grille[x][y].y = 30.0f-rand()%60;
		                     bezier.grille[x][y].z = y*SIZE - (SIZE * POINTSY)/2.0f+SIZE/2;
		               }   
	      }	     
	            
			if (id == resetButton->getId())
			{
			      for (int y=0; y<POINTSY; y++)
	                  for (int x=0; x<POINTSX; x++)
		               {
		                     bezier.grille[x][y].x = x*SIZE - (SIZE * POINTSX)/2.0f+SIZE/2;
		                     bezier.grille[x][y].y = 0.0f;
		                     bezier.grille[x][y].z = y*SIZE - (SIZE * POINTSY)/2.0f+SIZE/2;
		               }   
	      }
	      
	      if (id == hideButton->getId())
			{
			      if (hideButton->isClicked())
			      {
			            patch.affiche_pt_ctrl = false;
			            patch.affiche_vect_derive = false;
			      }
			      else
			      {
			            patch.affiche_pt_ctrl = true;
			            patch.affiche_vect_derive = true;
			      }			      
			}			           
			
			// Display:		
			MVISIO::swap();		
			Sleep(20); // Don't kill mthreading ;)
	}      
	
	// Release patches & co:
	for (int c=0; c<bezier.nb_lignes; c++)
         free(bezier.grille[c]);
	free(bezier.grille);					
   
   // Done:
   MVISIO::free();
   return 0;
}