/******************************************************************************/
/* TP : Cration de patch de bzier  partir de grille de points.             */          
/* Ocane LAGUERRE                                                            */
/* Renaud OTT                                                                 */
/******************************************************************************/

#include <math.h>
#ifdef WIN32
   #include <windows.h>
#endif
#include <GL/gl.h>
#include "bezier.h"



// La fonction qui calcule les points rsulatas  partir des points de controle
Grille_triplet calcul_patch_bezier(Grille_triplet Tab_pt_controle, int nb_point_u,
                                   int nb_point_v) {
	
	Grille_triplet liste;	// La grille rsultat renvoye par la fonction
	Grille_triplet tmp;		// une grille temporaire pour stocker les parties 
	                        // de Tab_pt_controle
	
	int truc, truc2;
	int num_lig;
	
	int iter = 0;
	int num_pt_ctrl = 0;

	int num_pt_u = 0;
	int num_pt_v = 0;
	double t_u, t_v;
	
	// On alloue le resultat
	liste.nb_lignes = nb_point_u;
	liste.nb_colonnes = nb_point_v;
	liste.grille = (Triplet **) malloc(sizeof(Triplet *) * liste.nb_lignes);
	for (truc = 0; truc < liste.nb_lignes; truc++)
		liste.grille[truc] = (Triplet *) malloc(sizeof(Triplet) * liste.nb_colonnes);
	
	// On alloue le tablerau temporaire	
	tmp.nb_lignes = Tab_pt_controle.nb_lignes;
	tmp.nb_colonnes = Tab_pt_controle.nb_colonnes;
	tmp.grille = (Triplet **) malloc(sizeof(Triplet *) * tmp.nb_lignes);
	for (truc = 0; truc < tmp.nb_lignes; truc++)
		tmp.grille[truc] = (Triplet *) malloc(sizeof(Triplet) * tmp.nb_colonnes);

	for (num_pt_u = 0; num_pt_u < nb_point_u; num_pt_u++) {
	
		t_u = ((double)num_pt_u)/((double)nb_point_u-1);
		for (num_pt_v = 0; num_pt_v < nb_point_v; num_pt_v++) {
			//Calcul de t_v
			t_v = ((double)num_pt_v)/((double)nb_point_v-1); 
		
			//On sauvegarde Grille_pt_controle
			for (truc = 0; truc < Tab_pt_controle.nb_lignes; truc++) {
				for (truc2 = 0; truc2 < Tab_pt_controle.nb_colonnes; truc2++) {		
					tmp.grille[truc][truc2].x = Tab_pt_controle.grille[truc][truc2].x;
					tmp.grille[truc][truc2].y = Tab_pt_controle.grille[truc][truc2].y;
					tmp.grille[truc][truc2].z = Tab_pt_controle.grille[truc][truc2].z;
				}
			}
		
			// On fait Casteljau sur toutes les lignes
			for (num_lig = 0; num_lig < tmp.nb_lignes; num_lig++) {
				for (iter = 0; iter < tmp.nb_colonnes; iter++) {
					for (num_pt_ctrl = 0; num_pt_ctrl < (tmp.nb_colonnes-iter-1);num_pt_ctrl++) {				
						tmp.grille[num_lig][num_pt_ctrl].x = tmp.grille[num_lig][num_pt_ctrl].x * (1.0-t_u) +
						                         tmp.grille[num_lig][num_pt_ctrl+1].x * (t_u);
						tmp.grille[num_lig][num_pt_ctrl].y = tmp.grille[num_lig][num_pt_ctrl].y * (1.0-t_u) +
						                         tmp.grille[num_lig][num_pt_ctrl+1].y * (t_u);
						tmp.grille[num_lig][num_pt_ctrl].z = tmp.grille[num_lig][num_pt_ctrl].z * (1.0-t_u) +
						                         tmp.grille[num_lig][num_pt_ctrl+1].z * (t_u);
					}
				}
			}
		
			// On fait Casteljau sur les colonnes
			for (iter = 0; iter < tmp.nb_lignes; iter++) {
				for (num_pt_ctrl = 0; num_pt_ctrl < (tmp.nb_lignes-iter-1);num_pt_ctrl++) {				 	 																								 	 
					tmp.grille[num_pt_ctrl][0].x = tmp.grille[num_pt_ctrl][0].x * (1.0-t_v) +	 
					                         tmp.grille[num_pt_ctrl+1][0].x * (t_v);				 	 
					tmp.grille[num_pt_ctrl][0].y = tmp.grille[num_pt_ctrl][0].y * (1.0-t_v) +	 
					                         tmp.grille[num_pt_ctrl+1][0].y * (t_v);				 	 	 
					tmp.grille[num_pt_ctrl][0].z = tmp.grille[num_pt_ctrl][0].z * (1.0-t_v) +
					                         tmp.grille[num_pt_ctrl+1][0].z * (t_v);				 	 	 
				}																						 	 		
			}
			
			// On affecte le rsultat  liste
			liste.grille[num_pt_u][num_pt_v].x = tmp.grille[0][0].x; 
			liste.grille[num_pt_u][num_pt_v].y = tmp.grille[0][0].y;
			liste.grille[num_pt_u][num_pt_v].z = tmp.grille[0][0].z;
		}
	}
	
	// On libre le tableau temporaire
	for (truc = 0; truc < tmp.nb_lignes; truc++)
		free(tmp.grille[truc]);
	free(tmp.grille);
	
	return liste;
} 


// La fonction qui recalcule tous les membres de la structure lorsqu'il y a changements
void changement(struct patch_bezier *o) {

	int i,j;
	int pars;
	double lg;
	int nb_pt_deriv_u;
 	int nb_pt_deriv_v;
	Grille_triplet gr_cont_deriv_u;
	Grille_triplet gr_result_deriv_u;
	Grille_triplet gr_cont_deriv_v;
	Grille_triplet gr_result_deriv_v;
		
	if (true) {
		
		// On recalcule le patch
		if (o->gr_result.nb_lignes) {	
			for (pars = 0; pars < o->gr_result.nb_lignes; pars++)
				free(o->gr_result.grille[pars]);
			free(o->gr_result.grille);
		}
		o->gr_result = calcul_patch_bezier(o->gr_cont, o->nb_pt_u, o->nb_pt_v);
	
		gr_cont_deriv_u.nb_lignes = o->gr_cont.nb_lignes - 1;
		gr_cont_deriv_u.nb_colonnes = o->gr_cont.nb_colonnes;
		gr_cont_deriv_v.nb_lignes = o->gr_cont.nb_lignes;
		gr_cont_deriv_v.nb_colonnes = o->gr_cont.nb_colonnes - 1;
		
		// Allocation des deux patches temporaires pour la derive
		gr_cont_deriv_u.grille = (Triplet **) malloc(sizeof(Triplet *) * o->gr_cont.nb_lignes-1);
		for (pars = 0; pars < gr_cont_deriv_u.nb_lignes; pars++)
			gr_cont_deriv_u.grille[pars] = (Triplet *) malloc(sizeof(Triplet) * gr_cont_deriv_u.nb_colonnes);
			
		gr_cont_deriv_v.grille = (Triplet **) malloc(sizeof(Triplet *) * o->gr_cont.nb_lignes);
		for (pars = 0; pars < gr_cont_deriv_v.nb_lignes; pars++)
			gr_cont_deriv_v.grille[pars] = (Triplet *) malloc(sizeof(Triplet) * gr_cont_deriv_v.nb_colonnes);
		
		// Initialisation des deux patchs temporaires pour la derive
		for (nb_pt_deriv_u = 0; nb_pt_deriv_u < gr_cont_deriv_u.nb_lignes; nb_pt_deriv_u++) {
			for (nb_pt_deriv_v = 0; nb_pt_deriv_v < gr_cont_deriv_u.nb_colonnes; nb_pt_deriv_v++) {
				gr_cont_deriv_u.grille[nb_pt_deriv_u][nb_pt_deriv_v].x =
										   o->gr_cont.grille[nb_pt_deriv_u + 1][nb_pt_deriv_v].x -
				 						   o->gr_cont.grille[nb_pt_deriv_u][nb_pt_deriv_v].x;
				 							  
				gr_cont_deriv_u.grille[nb_pt_deriv_u][nb_pt_deriv_v].y =
										   o->gr_cont.grille[nb_pt_deriv_u + 1][nb_pt_deriv_v].y -
				 						   o->gr_cont.grille[nb_pt_deriv_u][nb_pt_deriv_v].y;
				 						   
				gr_cont_deriv_u.grille[nb_pt_deriv_u][nb_pt_deriv_v].z =
				                           o->gr_cont.grille[nb_pt_deriv_u + 1][nb_pt_deriv_v].z -
				 						   o->gr_cont.grille[nb_pt_deriv_u][nb_pt_deriv_v].z;
												 
			}
		}
		for (nb_pt_deriv_u = 0; nb_pt_deriv_u < gr_cont_deriv_v.nb_lignes; nb_pt_deriv_u++) {
			for (nb_pt_deriv_v = 0; nb_pt_deriv_v < gr_cont_deriv_v.nb_colonnes; nb_pt_deriv_v++) {
				gr_cont_deriv_v.grille[nb_pt_deriv_u][nb_pt_deriv_v].x =
										   o->gr_cont.grille[nb_pt_deriv_u][nb_pt_deriv_v+1].x -
				 						   o->gr_cont.grille[nb_pt_deriv_u][nb_pt_deriv_v].x;
				 							  
				gr_cont_deriv_v.grille[nb_pt_deriv_u][nb_pt_deriv_v].y =
										   o->gr_cont.grille[nb_pt_deriv_u][nb_pt_deriv_v+1].y -
				 						   o->gr_cont.grille[nb_pt_deriv_u][nb_pt_deriv_v].y;
				 						   
				gr_cont_deriv_v.grille[nb_pt_deriv_u][nb_pt_deriv_v].z =
				                              o->gr_cont.grille[nb_pt_deriv_u][nb_pt_deriv_v+1].z -
				 							  o->gr_cont.grille[nb_pt_deriv_u][nb_pt_deriv_v].z;
												 
			}
		}
		
		// Calcul des deux derive
		gr_result_deriv_u = calcul_patch_bezier(gr_cont_deriv_u, o->nb_pt_u, o->nb_pt_v);
		gr_result_deriv_v = calcul_patch_bezier(gr_cont_deriv_v, o->nb_pt_u, o->nb_pt_v);
		
		// On calcule la drive suivant u et v
		// Libration 
		if (gr_cont_deriv_u.nb_lignes) {
			
			for (pars = 0; pars < gr_cont_deriv_u.nb_lignes; pars++)
				free(gr_cont_deriv_u.grille[pars]);
			free(gr_cont_deriv_u.grille);
		}
		if (gr_cont_deriv_v.nb_lignes > 0) {
			for (pars = 0; pars < gr_cont_deriv_v.nb_lignes; pars++)
				free(gr_cont_deriv_v.grille[pars]);
			free(gr_cont_deriv_v.grille);
		}

		if (o->gr_derivee.nb_lignes) {
			for (pars = 0; pars < o->gr_derivee.nb_lignes; pars++)
				free(o->gr_derivee.grille[pars]);
			free(o->gr_derivee.grille);
		}
		
		// On alloue et initialise la grille contenant les normales de chaque carreau du patch
		o->gr_derivee.nb_lignes = o->nb_pt_u - 1;
		o->gr_derivee.nb_colonnes = o->nb_pt_v - 1;
		o->gr_derivee.grille = (Triplet **) malloc(sizeof(Triplet *) * o->gr_derivee.nb_lignes);
		for (pars = 0; pars < o->gr_derivee.nb_lignes; pars++)
			o->gr_derivee.grille[pars] = (Triplet *) malloc(sizeof(Triplet) * o->gr_derivee.nb_colonnes);
		
		// On calcule les noramles en prenant bien soin de les rendre unitaires	
		for(i=0  ; i<o->nb_pt_u-1; i++) {
			for(j=0  ; j<o->nb_pt_v-1; j++) {
				// On calcule le vecteur derive
				o->gr_derivee.grille[i][j].x = gr_result_deriv_u.grille[i][j].y * gr_result_deriv_v.grille[i][j].z -
					                        gr_result_deriv_u.grille[i][j].z * gr_result_deriv_v.grille[i][j].y;
				o->gr_derivee.grille[i][j].y = gr_result_deriv_u.grille[i][j].z * gr_result_deriv_v.grille[i][j].x -
					                        gr_result_deriv_u.grille[i][j].x * gr_result_deriv_v.grille[i][j].z;
				o->gr_derivee.grille[i][j].z = gr_result_deriv_u.grille[i][j].x * gr_result_deriv_v.grille[i][j].y -
											gr_result_deriv_u.grille[i][j].y * gr_result_deriv_v.grille[i][j].x;
											
				// On normalise							
				lg = sqrt(pow(o->gr_derivee.grille[i][j].x, 2.0f) + pow(o->gr_derivee.grille[i][j].y, 2.0f) + 
				          pow(o->gr_derivee.grille[i][j].z, 2.0f));
				o->gr_derivee.grille[i][j].x /= lg;
				o->gr_derivee.grille[i][j].y /= lg;
				o->gr_derivee.grille[i][j].z /= lg;
			}	
		}
		
		// On libre les deux grilles resultat qui nous ont servi  calculer les vecteurs derivs
		if (gr_result_deriv_u.nb_lignes) {
			
			for (pars = 0; pars < gr_result_deriv_u.nb_lignes; pars++)
				free(gr_result_deriv_u.grille[pars]);
			free(gr_result_deriv_u.grille);
		}
		if (gr_result_deriv_v.nb_lignes > 0) {
			for (pars = 0; pars < gr_result_deriv_v.nb_lignes; pars++)
				free(gr_result_deriv_v.grille[pars]);
			free(gr_result_deriv_v.grille);
		} 
	}
}


// La fonction qui affiche le patch
void affiche_patch_bezier(struct patch_bezier *o, float blend)
{
	int j, i;
	
	// Les couleur de matire du patch par dfaut
	const float mat_ambient[] = { 0.25f, 0.25f, 0.0f, blend };
	const float mat_diffuse[] = { 0.85f, 0.85f, 0.0f, blend };
	const float mat_specular[] = { 0.45f, 0.45f, 0.0f, blend };
    const float mat_shininess[] = { 10.0f };
	float point_milieu[3];
                           
	glPushAttrib(GL_COLOR);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, mat_ambient);	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	float tempPosition[] = {0.0f, 100.0f, 0.0f, 1.0f};
	float tempDirection[] = {0.0f, 0.0f, 0.0f, 1.0f};
	 glLightf(GL_LIGHT0,  GL_SPOT_CUTOFF, 180.0f); 
	glLightfv(GL_LIGHT0, GL_POSITION, tempPosition);      
   glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, tempDirection);   
   float a[] = {0.1f, 0.1f, 0.1f, 1.0f};
   glLightfv(GL_LIGHT0, GL_AMBIENT, a);
   float d[] = {0.8f, 0.8f, 0.8f, 1.0f};
   glLightfv(GL_LIGHT0, GL_DIFFUSE, d);
   float s[] = {0.2f, 0.2f, 0.2f, 1.0f};
   glLightfv(GL_LIGHT0, GL_SPECULAR, s);
	glDisable(GL_CULL_FACE);	
	
	// affichage du patch solid:	
	for(i=0  ; i<o->nb_pt_u-1; i++) {
		for(j=0  ; j<o->nb_pt_v-1; j++) {
		
			glEnable(GL_LIGHTING);
			if (blend != 1.0f)
			      glEnable(GL_BLEND);
			//glColor4f(1.0f, 1.0f, 0.0f, 1.0f);
			// On spcifie la normale pour le rendu avec couleur et lumire
			glNormal3f(-o->gr_derivee.grille[i][j].x,
			           -o->gr_derivee.grille[i][j].y,
			           -o->gr_derivee.grille[i][j].z);
						
			// On dessine le carre
			glEnable(GL_POLYGON_OFFSET_FILL);
			glPolygonOffset(0.5, 0.5);
			
			glBegin(GL_POLYGON);
				glVertex3f(o->gr_result.grille[i][j].x,
				           o->gr_result.grille[i][j].y,
				           o->gr_result.grille[i][j].z);
				glVertex3f(o->gr_result.grille[i][j+1].x,
				           o->gr_result.grille[i][j+1].y,
				           o->gr_result.grille[i][j+1].z);
				glVertex3f(o->gr_result.grille[i+1][j+1].x,
				           o->gr_result.grille[i+1][j+1].y,
				           o->gr_result.grille[i+1][j+1].z);
				glVertex3f(o->gr_result.grille[i+1][j].x,
				           o->gr_result.grille[i+1][j].y,
				           o->gr_result.grille[i+1][j].z);
			glEnd();
			glDisable(GL_POLYGON_OFFSET_FILL);
			if (blend != 1.0f)
			      glDisable(GL_BLEND);	
	}
	}
	
	// affichage du patch wireframe	
	for(i=0  ; i<o->nb_pt_u-1; i++) {
		for(j=0  ; j<o->nb_pt_v-1; j++) {		   
			glDisable(GL_LIGHTING);
			glColor4f(0.7f, 0.7f, 0.1f, 1.0f);			
			
			//glColor4f(1.0f, 1.0f, 0.0f, 1.0f);
			// On spcifie la normale pour le rendu avec couleur et lumire
			glNormal3f(-o->gr_derivee.grille[i][j].x,
			           -o->gr_derivee.grille[i][j].y,
			           -o->gr_derivee.grille[i][j].z);
						
			// On dessine le carre
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);						
			glBegin(GL_POLYGON);
				glVertex3f(o->gr_result.grille[i][j].x,
				           o->gr_result.grille[i][j].y,
				           o->gr_result.grille[i][j].z);
				glVertex3f(o->gr_result.grille[i][j+1].x,
				           o->gr_result.grille[i][j+1].y,
				           o->gr_result.grille[i][j+1].z);
				glVertex3f(o->gr_result.grille[i+1][j+1].x,
				           o->gr_result.grille[i+1][j+1].y,
				           o->gr_result.grille[i+1][j+1].z);
				glVertex3f(o->gr_result.grille[i+1][j].x,
				           o->gr_result.grille[i+1][j].y,
				           o->gr_result.grille[i+1][j].z);
			glEnd();
			if (blend != 1.0f)
			      glDisable(GL_BLEND);						
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		
			// On dessine les vecteurs derivs si besoin			
			if (o->affiche_vect_derive) { 
				//On calcule le point milieu du carreau du patch
				point_milieu[0] = (o->gr_result.grille[i][j].x +  o->gr_result.grille[i][j+1].x + 
			                      o->gr_result.grille[i+1][j].x +  o->gr_result.grille[i+1][j+1].x)/4.0;
				
				point_milieu[1] = (o->gr_result.grille[i][j].y +  o->gr_result.grille[i][j+1].y + 
			                      o->gr_result.grille[i+1][j].y +  o->gr_result.grille[i+1][j+1].y)/4.0;
				
				point_milieu[2] = (o->gr_result.grille[i][j].z +  o->gr_result.grille[i][j+1].z +
				                  o->gr_result.grille[i+1][j].z +  o->gr_result.grille[i+1][j+1].z)/4.0; 
								  
				glDisable(GL_LIGHTING);				
				
				// On dessine la normale
				glBegin(GL_LINES);
				   glColor3f(0.0, 1.0, 0.2);
				   point_milieu[1] += 0.2f;
					glVertex3fv(point_milieu);
					glColor3f(0.0, 0.8, 0.0);
					glVertex3f(point_milieu[0] - o->gr_derivee.grille[i][j].x/0.5,
	    			           point_milieu[1] - o->gr_derivee.grille[i][j].y/0.5,
	    			           point_milieu[2] - o->gr_derivee.grille[i][j].z/0.5);
				glEnd();
				
			}
		}	
	}	
	// affichage du polygone de controle	
	if (o->affiche_pt_ctrl) {
	    glDisable(GL_LIGHTING);
	    
		glLineWidth(3.0);		
 		glColor3f(1.0, 0.0, 0.0);
 		
		for(i=0  ; i<o->gr_cont.nb_lignes-1; i++) {
			for(j=0  ; j<o->gr_cont.nb_colonnes-1; j++) {
			glBegin(GL_LINE_LOOP);
			glVertex3f(o->gr_cont.grille[i][j].x,
	    		 o->gr_cont.grille[i][j].y,
	    		 o->gr_cont.grille[i][j].z);
			glVertex3f(o->gr_cont.grille[i+1][j].x,
	    		 o->gr_cont.grille[i+1][j].y,
	    		 o->gr_cont.grille[i+1][j].z);
			glVertex3f(o->gr_cont.grille[i+1][j+1].x,
	    		 o->gr_cont.grille[i+1][j+1].y,
	    		 o->gr_cont.grille[i+1][j+1].z);
			glVertex3f(o->gr_cont.grille[i][j+1].x,
	    		 o->gr_cont.grille[i][j+1].y,
	    		 o->gr_cont.grille[i][j+1].z);
			glEnd();
			}
		}
		glLineWidth(1.0);
	} 		
  
   glDisable(GL_LIGHTING);
   glDisable(GL_LIGHT0);
  	glPopAttrib();
}
